@@ -2,24 +2,25 @@
 {-# LANGUAGE QuasiQuotes                #-}
 {-# LANGUAGE ConstraintKinds            #-}
 {-# LANGUAGE FlexibleContexts           #-}
+{-# LANGUAGE FlexibleInstances          #-}
 {-# LANGUAGE TypeFamilies               #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
+{-# LANGUAGE MultiParamTypeClasses      #-}
 {-# LANGUAGE TemplateHaskell            #-}
 {-# LANGUAGE OverloadedStrings          #-}
 {-# LANGUAGE GADTs                      #-}
-{-# LANGUAGE CPP                        #-}
 -------------------------------------------------------------------------------
 -- |
 -- Module      :  Yesod.Auth.HashDB
--- Copyright   :  (c) Patrick Brisbin 2010, Paul Rouse 2014
+-- Copyright   :  (c) Patrick Brisbin 2010, Paul Rouse 2014-2015
 -- License     :  MIT
 --
 -- Maintainer  :  Paul Rouse <pyr@doynton.org>
 -- Stability   :  Stable
 -- Portability :  Portable
 --
--- A yesod-auth AuthPlugin designed to look users up in Persist where
--- their user id's and a hash of their password is stored.
+-- A yesod-auth AuthPlugin designed to look users up in a Persistent
+-- database where the hash of their password is stored.
 --
 -- This module was removed from @yesod-auth-1.3.0.0@ and is now
 -- maintained separately.
@@ -50,51 +51,65 @@
 -- and can ignore the salt.
 --
 -- In a system which has been migrated from the old format, passwords
--- which are reset using the new format will have an empty salt field.
+-- which are reset will use the new format and will have an empty salt field.
 -- Once all the entries are of this form, it is safe to change the model
 -- to remove the salt, and change the 'HashDBUser' instance accordingly.
 --
--- To use this in a Yesod application, it must be an instance of
--- YesodPersist, and the username and hashed-passwords should be added
--- to the database.  The followng steps give an outline of what is required.
+-- To use this in a Yesod application, the foundation data type must be an
+-- instance of YesodPersist, and the username and hashed passwords should
+-- be added to the database.  The following steps give an outline of what
+-- is required.
 --
 -- You need a database table to store user records: in a scaffolded site it
 -- might look like:
 --
 -- > User
--- >     name Text             -- user name used by HashDB
+-- >     name Text             -- user name used to uniquely identify users
 -- >     password Text Maybe   -- password hash for HashDB
 -- >     UniqueUser name
 --
--- Create an instance of 'HashDBUser' for this data type:
+-- Create an instance of 'HashDBUser' for this data type.  For historical
+-- reasons "Yesod.Auth.HashDB" exports some names which are quite likely to
+-- clash with your own, so it is a good idea to import just the ones you need:
 --
+-- > import Yesod.Auth.HashDB (HashDBUser(..))
+-- > ....
 -- > instance HashDBUser User where
 -- >     userPasswordHash = userPassword
--- >     setPasswordHash h p = p { userPassword = Just h }
+-- >     setPasswordHash h u = u { userPassword = Just h }
 --
 -- In the YesodAuth instance declaration for your app, include 'authHashDB'
 -- like so:
 --
+-- > import Yesod.Auth.HashDB (authHashDB)
+-- > ....
 -- > instance YesodAuth App where
 -- >     ....
 -- >     authPlugins _ = [ authHashDB (Just . UniqueUser), .... ]
--- >     getAuthId = getAuthIdHashDB AuthR (Just . UniqueUser)
+-- >     getAuthId creds = ... -- Perhaps modify scaffolding: see below
 --
--- @AuthR@ should be your authentication route, and the function
--- @(Just . UniqueUser)@ supplied to both 'authHashDB' and
--- 'getAuthIdHashDB' takes a 'Text' and produces a 'Unique' value to
--- look up in the User table.  'getAuthIdHashDB' is just a convenience
--- for the case when 'HashDB' is the only plugin, and something else
--- would be needed when other plugins are used as well.
+-- The argument to 'authHashDB' is a function which takes a 'Text' and
+-- produces a 'Maybe' containing a 'Unique' value to look up in the User
+-- table.  The example @(Just . UniqueUser)@ shown here works for the
+-- model outlined above.
 --
--- You can create password hashes manually as follows, if you need to
--- initialise the database:
+-- In the scaffolding, the definition of @getAuthId@ contains code to
+-- add a user who is not already in the database.  Depending on how users
+-- are administered, this may not make sense when using HashDB, so consider
+-- whether it should be removed.
+--
+-- For a real application, the developer should provide some sort of
+-- of administrative interface for setting passwords; it needs to call
+-- 'setPassword' and save the result in the database.  However, if you
+-- need to initialise the database by hand, you can generate the correct
+-- password hash as follows:
 --
 -- > ghci -XOverloadedStrings
 -- > > import Crypto.PasswordStore
--- > > makePassword "MyPassword" 14
+-- > > makePassword "MyPassword" 17
 --
--- where \"14\" is the default strength parameter used in this module.
+-- where \"17\" is the default strength parameter ('defaultStrength') used
+-- in this module.
 --
 -- == Custom Login Form
 --
@@ -128,17 +143,16 @@ module Yesod.Auth.HashDB
     , defaultStrength
     , setPasswordStrength
     , setPassword
+    , validatePass
     , upgradePasswordHash
-      -- * Authentication
+      -- * Interface to database and Yesod.Auth
     , validateUser
     , authHashDB
     , authHashDBWithForm
     , getAuthIdHashDB
       -- * Predefined data type
     , User
-#if !MIN_VERSION_persistent(2, 0, 0)
     , UserGeneric (..)
-#endif
     , UserId
     , EntityField (..)
     , migrateUsers
@@ -148,6 +162,7 @@ import Yesod.Persist
 import Yesod.Form
 import Yesod.Auth
 import Yesod.Core
+import qualified Yesod.Auth.Message as Msg
 
 import Control.Applicative         ((<$>), (<*>))
 import Data.Typeable
@@ -162,10 +177,11 @@ import Crypto.PasswordStore        (makePassword, verifyPassword,
 -- | Default strength used for passwords (see "Crypto.PasswordStore" for
 --   details).
 defaultStrength :: Int
-defaultStrength = 14
+defaultStrength = 17
 
--- | Interface for data type which holds user info. It's just a
---   collection of getters and setters
+-- | The type representing user information stored in the database should
+--   be an instance of this class.  It just provides the getters and setters
+--   used by the functions in this module.
 class HashDBUser user where
   -- | Retrieve password hash from user data
   userPasswordHash :: user -> Maybe Text
@@ -208,7 +224,9 @@ class HashDBUser user where
   setSaltAndPasswordHash = setUserHashAndSalt
 
   {-# MINIMAL userPasswordHash, (setPasswordHash | (userPasswordSalt, setSaltAndPasswordHash)) #-}
+{-# DEPRECATED userPasswordSalt "Compatibility with old data containing a separate salt field will be removed eventually" #-}
 {-# DEPRECATED setUserHashAndSalt "Please use setSaltAndPasswordHash instead" #-}
+{-# DEPRECATED setSaltAndPasswordHash "Compatibility with old data containing a separate salt field will be removed eventually" #-}
 
 
 -- | Calculate salted hash using SHA1.  Retained for compatibility with
@@ -242,6 +260,42 @@ setPasswordStrength strength pwd u = do
 setPassword :: (MonadIO m, HashDBUser user) => Text -> user -> m user
 setPassword = setPasswordStrength defaultStrength
 
+-- | Validate a plaintext password against the hash in the user data structure.
+--   This function retains compatibility with user data produced by old
+--   versions of this module (prior to 1.3), although the hashes are less
+--   secure and should be upgraded as soon as possible.  They can be
+--   upgraded using 'upgradePasswordHash', or by insisting that users set
+--   new passwords.
+--
+--   The result distinguishes two types of validation failure, which may
+--   be useful in an application which supports multiple authentication
+--   methods:
+--
+--   * Just False - the user has a password set up, but the given one does
+--     not match it
+--
+--   * Nothing - the user does not have a password (the hash is Nothing)
+--
+--   Since 1.4.1
+--
+validatePass :: HashDBUser u => u -> Text -> Maybe Bool
+validatePass user passwd = do
+    hash <- userPasswordHash user
+    salt <- userPasswordSalt user
+    -- NB plaintext password characters are truncated to 8 bits here, and also
+    -- in saltedHash and passwordHash above (the hash and old salt are already
+    -- 8 bit).  This is for historical compatibility, but in practice it is
+    -- unlikely to reduce the entropy of most users' alphabets by much.
+    let hash' = BS.pack $ unpack hash
+        passwd' = BS.pack $ unpack $ if salt == "" then passwd
+                                     else
+                                         -- Extra layer for an upgraded old hash
+                                         saltedHash salt passwd
+    if passwordStrength hash' > 0
+        -- Will give >0 for new-style hash, else fall back
+        then return $ verifyPassword passwd' hash'
+        else return $ hash == saltedHash salt passwd
+
 -- | Upgrade existing user credentials to a stronger hash.  The existing
 --   hash may have been produced either by previous versions of this module,
 --   which used a weak algorithm, or from a weaker setting in the current
@@ -281,54 +335,33 @@ upgradePasswordHash strength u = do
 
 
 ----------------------------------------------------------------
--- Authentication
+-- Interface to database and Yesod.Auth
 ----------------------------------------------------------------
 
--- | Constraint for types of functions in this module
+-- | Constraint for types of interface functions in this module
 --
-#if MIN_VERSION_persistent(2, 0, 0)
 type HashDBPersist master user =
     ( YesodAuthPersist master
     , PersistUnique (YesodPersistBackend master)
     , AuthEntity master ~ user
+    , AuthId master ~ Key user
+    , PersistEntityBackend user ~ YesodPersistBackend master
     , HashDBUser user
+    , PersistEntity user
     )
-#else
-type HashDBPersist master user =
-    ( YesodAuthPersist master
-    , PersistUnique (YesodPersistBackend master (HandlerT master IO))
-    , AuthEntity master ~ user
-    , HashDBUser user
-    )
-#endif
 
 -- | Given a user ID and password in plaintext, validate them against
---   the database values.  This function retains compatibility with
---   databases containing hashes produced by previous versions of this
---   module, although they are less secure and should be upgraded as
---   soon as possible.  They can be upgraded using 'upgradePasswordHash',
---   or by insisting that users set new passwords.
+--   the database values.  This function simply looks up the user id in the
+--   database and calls 'validatePass' to do the work.
+--
 validateUser :: HashDBPersist site user =>
                 Unique user     -- ^ User unique identifier
              -> Text            -- ^ Password in plaintext
              -> HandlerT site IO Bool
 validateUser userID passwd = do
-  -- Checks that hash and password match
-  let validate u = do hash <- userPasswordHash u
-                      salt <- userPasswordSalt u
-                      let hash' = BS.pack $ unpack hash
-                          passwd' = BS.pack $ unpack
-                              $ if salt == "" then passwd
-                                else
-                                    -- Extra layer for an upgraded old hash
-                                    saltedHash salt passwd
-                      if passwordStrength hash' > 0
-                        -- Will give >0 for new-style hash, else fall back
-                        then return $ verifyPassword passwd' hash'
-                        else return $ hash == saltedHash salt passwd
   -- Get user data
   user <- runDB $ getBy userID
-  return $ fromMaybe False $ validate . entityVal =<< user
+  return $ fromMaybe False $ flip validatePass passwd . entityVal =<< user
 
 
 login :: AuthRoute
@@ -348,10 +381,8 @@ postLoginR uniq = do
     isValid <- lift $ fromMaybe (return False) 
                  (validateUser <$> (uniq =<< mu) <*> mp)
     if isValid 
-       then lift $ setCredsRedirect $ Creds "hashdb" (fromMaybe "" mu) []
-       else do
-           tm <- getRouteToParent
-           lift $ loginErrorMessage (tm LoginR) "Invalid username/password"
+        then lift $ setCredsRedirect $ Creds "hashdb" (fromMaybe "" mu) []
+        else loginErrorMessageI LoginR Msg.InvalidUsernamePass
 
 
 -- | A drop in for the getAuthId method of your YesodAuth instance which
@@ -374,8 +405,11 @@ getAuthIdHashDB authR uniq creds = do
                 -- user exists
                 Just (Entity uid _) -> return $ Just uid
                 Nothing       -> do
-                                     _ <- loginErrorMessage (authR LoginR) "User not found"
-                                     return Nothing
+                    mr <- getMessageRender
+                    _ <- loginErrorMessage (authR LoginR) (mr Msg.InvalidUsernamePass)
+                    return Nothing
+
+{-# DEPRECATED getAuthIdHashDB "If this is a problem, please discuss at <https://github.com/paul-rouse/yesod-auth-hashdb/issues/5>" #-}
 
 -- | Prompt for username and password, validate that against a database
 --   which holds the username and a hash of the password
@@ -442,7 +476,7 @@ $newline never
 ----------------------------------------------------------------
 
 -- | Generate data base instances for a valid user
-share [mkPersist sqlSettings, mkMigrate "migrateUsers"]
+share [mkPersist sqlSettings { mpsGeneric = True }, mkMigrate "migrateUsers"]
          [persistUpperCase|
 User
     username Text Eq
@@ -453,11 +487,7 @@ User
 |]
 {-# DEPRECATED User, migrateUsers "The predefined User data type will be removed soon - please define your own database table and accompanying instance of HashDBUser" #-}
 
-#if MIN_VERSION_persistent(2, 0, 0)
-instance HashDBUser User where
-#else
 instance HashDBUser (UserGeneric backend) where
-#endif
   userPasswordHash = Just . userPassword
   userPasswordSalt = Just . userSalt
   setSaltAndPasswordHash s h u = u { userSalt     = s
